/*
 * testSwStateCyclic.cpp
 *
 *  Created on: 2022/06/02
 *      Author: 
 */

// https://github.com/meekrosoft/fff
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <pthread.h>

// Want to keep the argument history for 512 calls
#define FFF_ARG_HISTORY_LEN 512
// Want to keep the call sequence history for 512 function calls
#define FFF_CALL_HISTORY_LEN 512

#include "fff/gtest/gtest.h"
#include "fff/fff.h"
DEFINE_FFF_GLOBALS;

void* gp_blk = (void*)NULL;			// 送信するメモリプールのアドレスの確認用

long loop_num;	 						// 無限ループ実行回数
jmp_buf buf; 							// 無限ループ脱出用のスタック退避領域

extern pthread_mutex_t mutex;			// テストケースを逐次実行するための保険

extern "C" {
	// テスト対象のCソースファイルを丸ごとインクルードすることにより、
	// static関数やstatic変数の参照・更新が可能となる。
	#include "../sw_state_cyclic.c"

}

/* 関数の偽装宣言 */
/* 関数の実体があるものはテスト対象のCソースファイルに記載の関数のみです。 */
/* 実体のない関数は、FFFで偽装します。 */
/* サービスコールの偽装宣言 */
/* タスク付属同期機能 */
FAKE_VALUE_FUNC(ER, slp_tsk);
FAKE_VALUE_FUNC(ER, iwup_tsk, ID);
/* 同期・通信機能 */
FAKE_VALUE_FUNC(ER, psnd_dtq , ID , intptr_t);
/* メモリプール管理機能 */
FAKE_VALUE_FUNC(ER, get_mpf, ID , void**);

/* ライブラリの偽装宣言 */
FAKE_VOID_FUNC(t_perror ,uint_t ,  const char* , int_t , const char* ,ER );
FAKE_VOID_FUNC(set_led2_state, unsigned char);

FAKE_VALUE_FUNC(unsigned char, get_sw1_state);
FAKE_VALUE_FUNC(unsigned char, get_push1_state);


ER slp_tsk_custom_fake(void)
{
	if(1 < slp_tsk_fake.call_count){
		printf("偽装slp_tak経由で無限ループ脱出!\n");
		longjmp(buf,1);
		printf( "無限ループ脱出したのでここは実行されない\n" );
	}
	return(E_OK);

}

// カスタムFake関数の実装
ER get_mpf_custom_fake(ID mpfid, void **p_blk)
{
	*p_blk = malloc(sizeof(64));					// コンフィギュレータで指定する固定長メモリプールのサイズ

	gp_blk = *p_blk;								// 送信するアドレスの検証用に保持

	return(E_OK);
}

intptr_t g_event[2];
int      gi_event;

ER psnd_dtq_custom_fake(ID dtqid, intptr_t data)
{
	EXPECT_EQ(gp_blk, (void *)data);	     // 送信するアドレスの検証
	g_event[gi_event] = ((T_EVENT *)data) -> event;    // メッセージのIDを保持

	gi_event ++;

	free((void*)data);



	return(E_OK);
}

unsigned char gret_sw;

unsigned char get_sw1_state_custom_fake(void)
{
	return(gret_sw);
}


unsigned char gret_push;

unsigned char get_push1_state_custom_fake(void)
{
	return(gret_push);
}


/* List of fakes used by this unit tester */
#define FFF_FAKES_LIST(FAKE)	\
    	FAKE(slp_tsk)			\
        FAKE(iwup_tsk)			\
		FAKE(get_mpf)			\
		FAKE(get_sw1_state)     \
		FAKE(get_push1_state)   \
		FAKE(set_led2_state)    \
		FAKE(slp_tsk)           \
		FAKE(psnd_dtq)

/* テストケースの初期化と後処理 									*/
/* テスト対象、及びテストケースはグローバル変数を使用している。 	*/
/* マルチスレッド環境下で複数のテストケースが同時実行された場合、	*/
/* グローバル変数のアクセスでテストが破綻する。  					*/
/* 複数のテストケースを同時実行することは無いとは思うが、 			*/
/* 保険としてmutexにてテストケースを逐次実行するようにした。		*/

class testSwStateCyclic: public testing::Test {
	  void SetUp() {
    	pthread_mutex_lock(&mutex);
    	/* Register resets */
    	FFF_FAKES_LIST(RESET_FAKE);
    	/* reset common FFF internal structures */
        FFF_RESET_HISTORY();
    }
    void TearDown() {
	    pthread_mutex_unlock(&mutex);
    }
};

// 割込みハンドラのテスト
TEST_F(testSwStateCyclic, testSwScanHandler)
{
	sw_scan_handler(0);/*テスト対象をコール*/

	//iwup_tskがコールされていることを確認
	EXPECT_EQ(fff.call_history[0],(void*)iwup_tsk);
	// iwup_tskの引数が正しいことを確認
	EXPECT_EQ(SW_SCAN_TASK,iwup_tsk_fake.arg0_history[0]);
}


TEST_F(testSwStateCyclic, testSwScanTaskSw1on)
{
	get_mpf_fake.custom_fake = get_mpf_custom_fake; 	// get_mpfをカスタムフェイク関数に設定
	psnd_dtq_fake.custom_fake = psnd_dtq_custom_fake;	// psnd_dtqをカスタムフェイク関数に設定
	get_sw1_state_fake.custom_fake = get_sw1_state_custom_fake;
	get_push1_state_fake.custom_fake = get_push1_state_custom_fake;
	slp_tsk_fake.custom_fake = slp_tsk_custom_fake;



	sw1 = OFF;
	sw8 = OFF;
	gret_sw = ON;
	gret_push = OFF;
	gi_event = 0;

	if(0 == setjmp(buf)){
		sw_scan_task(0);
	}
	else {
		printf( "無限ループ脱出成功 \n\n" );
	}

	EXPECT_EQ(EV_TIMER_SW1_ON, g_event[0]);

	// get_mpfがコールされていることを確認
	EXPECT_EQ(fff.call_history[2],(void*)get_mpf);
	// get_mpfの第1引数が正しいことを確認
	EXPECT_EQ(EVENT_MPF,get_mpf_fake.arg0_history[0]);

	EXPECT_EQ(ON,(unsigned char)sw1);


}


TEST_F(testSwStateCyclic, testSwScanTaskSw1off)
{
	get_mpf_fake.custom_fake = get_mpf_custom_fake; 	// get_mpfをカスタムフェイク関数に設定
	psnd_dtq_fake.custom_fake = psnd_dtq_custom_fake;	// psnd_dtqをカスタムフェイク関数に設定
	get_sw1_state_fake.custom_fake = get_sw1_state_custom_fake;
	get_push1_state_fake.custom_fake = get_push1_state_custom_fake;
	slp_tsk_fake.custom_fake = slp_tsk_custom_fake;



	sw1 = ON;
	sw8 = OFF;
	gret_sw = OFF;
	gret_push = OFF;
	gi_event = 0;

	if(0 == setjmp(buf)){
		sw_scan_task(0);						// blink点滅周期タスクのメインルーチン(無限ループ)へ突入
	}
	else {
		printf( "無限ループ脱出成功 \n\n" );
	}

	EXPECT_EQ(EV_TIMER_SW1_OFF,g_event[1]);

	// get_mpfがコールされていることを確認
	EXPECT_EQ(fff.call_history[2],(void*)get_mpf);

	// get_mpfの第1引数が正しいことを確認
	EXPECT_EQ(EVENT_MPF,get_mpf_fake.arg0_history[0]);

	EXPECT_EQ(OFF,(unsigned char)sw1);


	// 第2引数はローカル変数へのポインタのため、ここでは確認できない。
	// psnd_dtq_custom_fake内でメッセージのメンバを確認する
}

TEST_F(testSwStateCyclic, testSwScanTaskPushon)
{
	get_mpf_fake.custom_fake = get_mpf_custom_fake; 	// get_mpfをカスタムフェイク関数に設定
	psnd_dtq_fake.custom_fake = psnd_dtq_custom_fake;	// psnd_dtqをカスタムフェイク関数に設定
	get_sw1_state_fake.custom_fake = get_sw1_state_custom_fake;
	get_push1_state_fake.custom_fake = get_push1_state_custom_fake;
	slp_tsk_fake.custom_fake = slp_tsk_custom_fake;


	sw1 = ON;
	sw8 = OFF;
	gret_sw = ON;
	gret_push = ON;
	gi_event = 0;

	if(0 == setjmp(buf)){
		sw_scan_task(0);
	}
	else {
		printf( "無限ループ脱出成功 \n\n" );
	}

	EXPECT_EQ(EV_TIMER_PUSH1_ON,g_event[0]);

	// get_mpfがコールされていることを確認
	EXPECT_EQ(fff.call_history[3],(void*)get_mpf);
	// get_mpfの第1引数が正しいことを確認
	EXPECT_EQ(EVENT_MPF,get_mpf_fake.arg0_history[0]);


	EXPECT_EQ(ON,(unsigned char)sw8);

}


TEST_F(testSwStateCyclic, testSwScanTaskPushoff)
{
	get_mpf_fake.custom_fake = get_mpf_custom_fake; 	// get_mpfをカスタムフェイク関数に設定
	psnd_dtq_fake.custom_fake = psnd_dtq_custom_fake;	// psnd_dtqをカスタムフェイク関数に設定
	get_sw1_state_fake.custom_fake = get_sw1_state_custom_fake;
	get_push1_state_fake.custom_fake = get_push1_state_custom_fake;
	slp_tsk_fake.custom_fake = slp_tsk_custom_fake;



	sw1 = ON;
	sw8 = ON;
	gret_sw = ON;
	gret_push = OFF;
	if(0 == setjmp(buf)){
		sw_scan_task(0);
	}
	else {
		printf( "無限ループ脱出成功 \n\n" );
	}


	EXPECT_EQ(OFF,(unsigned char)sw8);

}


TEST_F(testSwStateCyclic, testSwScanTaskDouble)/*Sw1onPushon*/
{
	get_mpf_fake.custom_fake = get_mpf_custom_fake; 	// get_mpfをカスタムフェイク関数に設定
	psnd_dtq_fake.custom_fake = psnd_dtq_custom_fake;	// psnd_dtqをカスタムフェイク関数に設定
	get_sw1_state_fake.custom_fake = get_sw1_state_custom_fake;
	get_push1_state_fake.custom_fake = get_push1_state_custom_fake;
	slp_tsk_fake.custom_fake = slp_tsk_custom_fake;

	sw1 = OFF;
	sw8 = OFF;
	gret_sw = ON;
	gret_push = ON;
	gi_event = 0;

	if(0 == setjmp(buf)){
		sw_scan_task(0);
	}
	else {
		printf( "無限ループ脱出成功 \n\n" );
	}
	   EXPECT_EQ(EV_TIMER_SW1_ON, g_event[0]);/*sw1on*/

	   // get_mpfがコールされていることを確認
	   EXPECT_EQ(fff.call_history[2],(void*)get_mpf);
	   // get_mpfの第1引数が正しいことを確認
	   EXPECT_EQ(EVENT_MPF,get_mpf_fake.arg0_history[0]);

	   EXPECT_EQ(ON,(unsigned char)sw1);

	   EXPECT_EQ(EV_TIMER_PUSH1_ON,g_event[1]); /*pushon*/

	   // get_mpfがコールされていることを確認
	   EXPECT_EQ(fff.call_history[5],(void*)get_mpf);
	   // get_mpfの第1引数が正しいことを確認
	   EXPECT_EQ(EVENT_MPF,get_mpf_fake.arg0_history[0]);
	   EXPECT_EQ(ON,(unsigned char)sw8);
}






